@@ -1,4 +1,4 @@
-from typing import Any, Callable, Optional, Sequence, Type, TypeVar
+from typing import Any, Callable, Optional, Sequence, Type, TypeVar, Union
 
 import torch
 from einops import pack, rearrange, repeat, unpack
@@ -7,20 +7,27 @@ from typing_extensions import TypeGuard
 
 V = TypeVar("V")
 
+"""
+Helper functions
+"""
+
 
 def T(t: Callable[..., V]) -> Callable[..., Callable[..., V]]:
     """Where the magic happens, builds a type template for a given type"""
     return lambda *a, **ka: lambda *b, **kb: t(*(*a, *b), **{**ka, **kb})
 
 
+def Ts(t: Callable[..., V]) -> Callable[..., Callable[..., V]]:
+    """Builds a type template for a given type that accepts a list of instances"""
+    return lambda *types: lambda: t(*[tp() for tp in types])
+
+
 def exists(val: Optional[V]) -> TypeGuard[V]:
     return val is not None
 
 
 def default(val: Optional[V], d: V) -> V:
-    if exists(val):
-        return val
-    return d
+    return val if exists(val) else d
 
 
 def Module(modules: Sequence[nn.Module], forward_fn: Callable):
@@ -38,13 +45,11 @@ def Module(modules: Sequence[nn.Module], forward_fn: Callable):
 
 
 class Sequential(nn.Module):
-    """Custom Sequential that forwards args"""
+    """Custom Sequential that includes all args"""
 
     def __init__(self, *blocks):
         super().__init__()
-        self.blocks = nn.ModuleList(
-            [b if isinstance(b, nn.Module) else b() for b in blocks]
-        )
+        self.blocks = nn.ModuleList(blocks)
 
     def forward(self, x: Tensor, *args) -> Tensor:
         for block in self.blocks:
@@ -53,7 +58,7 @@ class Sequential(nn.Module):
 
 
 def Select(args_fn: Callable) -> Callable[..., Type[nn.Module]]:
-    """Selects (swap, remove) forward arguments given a (lambda) function"""
+    """Selects (swap, remove, repeat) forward arguments given a (lambda) function"""
 
     def fn(block_t: Type[nn.Module]) -> Type[nn.Module]:
         class Select(nn.Module):
@@ -82,8 +87,9 @@ class Packed(Sequential):
         return x
 
 
-def Repeat(block: Any, times: int) -> Any:
-    return Sequential(*(block,) * times)
+def Repeat(m: Union[nn.Module, Type[nn.Module]], times: int) -> Any:
+    ms = (m,) * times
+    return Sequential(*ms) if isinstance(m, nn.Module) else Ts(Sequential)(*ms)
 
 
 def Skip(merge_fn: Callable[[Tensor, Tensor], Tensor] = torch.add) -> Type[Sequential]:
@@ -97,24 +103,30 @@ def Skip(merge_fn: Callable[[Tensor, Tensor], Tensor] = torch.add) -> Type[Seque
     return Skip
 
 
+"""
+Modules
+"""
+
+
 def Conv(dim: int, *args, **kwargs) -> nn.Module:
     return [nn.Conv1d, nn.Conv2d, nn.Conv3d][dim - 1](*args, **kwargs)
 
 
-def Downsample(factor: int = 2, conv_t=Conv, **kwargs) -> nn.Module:
-    return conv_t(kernel_size=factor, stride=factor, **kwargs)
+def Downsample(dim: int, factor: int = 2, conv_t=Conv, **kwargs) -> nn.Module:
+    return conv_t(dim=dim, kernel_size=factor, stride=factor, **kwargs)
 
 
 def Upsample(
-    factor: int = 2, mode: str = "nearest", conv_t=Conv, **kwargs
+    dim: int, factor: int = 2, mode: str = "nearest", conv_t=Conv, **kwargs
 ) -> nn.Module:
     return nn.Sequential(
         nn.Upsample(scale_factor=factor, mode="nearest"),
-        conv_t(kernel_size=3, padding=1, **kwargs),
+        conv_t(dim=dim, kernel_size=3, padding=1, **kwargs),
     )
 
 
 def ConvBlock(
+    dim: int,
     in_channels: int,
     activation_t=nn.SiLU,
     norm_t=T(nn.GroupNorm)(num_groups=1),
@@ -124,14 +136,14 @@ def ConvBlock(
     return nn.Sequential(
         norm_t(num_channels=in_channels),
         activation_t(),
-        conv_t(in_channels=in_channels, **kwargs),
+        conv_t(dim=dim, in_channels=in_channels, **kwargs),
     )
 
 
 def ResnetBlock(
+    dim: int,
     in_channels: int,
     out_channels: int,
-    dim: Optional[int] = None,
     conv_block_t=ConvBlock,
     conv_t=Conv,
 ) -> nn.Module:
@@ -149,18 +161,7 @@ def ResnetBlock(
     return Module([conv_block, conv], lambda x: conv_block(x) + conv(x))
 
 
-def add_mask(sim: Tensor, mask: Tensor) -> Tensor:
-    b, ndim = sim.shape[0], mask.ndim
-    if ndim == 3:
-        mask = rearrange(mask, "b n m -> b 1 n m")
-    if ndim == 2:
-        mask = repeat(mask, "n m -> b 1 n m", b=b)
-    max_neg_value = -torch.finfo(sim.dtype).max
-    sim = sim.masked_fill(~mask, max_neg_value)
-    return sim
-
-
-def AttentionBase(features: int, head_features: int, num_heads: int):
+def AttentionBase(features: int, head_features: int, num_heads: int) -> nn.Module:
     scale = head_features**-0.5
     mid_features = head_features * num_heads
     to_out = nn.Linear(in_features=mid_features, out_features=features, bias=False)
@@ -173,9 +174,8 @@ def AttentionBase(features: int, head_features: int, num_heads: int):
         q, k, v = map(lambda t: rearrange(t, "b n (h d) -> b h n d", h=h), (q, k, v))
         # Compute similarity matrix and add eventual mask
         sim = einsum("... n d, ... m d -> ... n m", q, k) * scale
-        sim = add_mask(sim, mask) if exists(mask) else sim
         # Get attention matrix with softmax
-        attn = sim.softmax(dim=-1, dtype=torch.float32)
+        attn = sim.softmax(dim=-1)
         # Compute values
         out = einsum("... n m, ... m d -> ... n d", attn, v)
         out = rearrange(out, "b h n d -> b n (h d)")
@@ -184,10 +184,31 @@ def AttentionBase(features: int, head_features: int, num_heads: int):
     return Module([to_out], forward)
 
 
+def LinearAttentionBase(features: int, head_features: int, num_heads: int) -> nn.Module:
+    scale = head_features**-0.5
+    mid_features = head_features * num_heads
+    to_out = nn.Linear(in_features=mid_features, out_features=features, bias=False)
+
+    def forward(q: Tensor, k: Tensor, v: Tensor) -> Tensor:
+        h = num_heads
+        # Split heads
+        q, k, v = map(lambda t: rearrange(t, "b n (h d) -> b h n d", h=h), (q, k, v))
+        # Softmax rows and cols
+        q = q.softmax(dim=-1) * scale
+        k = k.softmax(dim=-2)
+        # Attend on channel dim
+        attn = einsum("... n d, ... n c -> ... d c", k, v)
+        out = einsum("... n d, ... d c -> ... n c", q, attn)
+        out = rearrange(out, "b h n d -> b n (h d)")
+        return to_out(out)
+
+    return Module([to_out], forward)
+
+
 def FixedEmbedding(max_length: int, features: int):
     embedding = nn.Embedding(max_length, features)
 
-    def forward(self, x: Tensor) -> Tensor:
+    def forward(x: Tensor) -> Tensor:
         batch_size, length, device = *x.shape[0:2], x.device
         assert_message = "Input sequence length must be <= max_length"
         assert length <= max_length, assert_message
@@ -261,10 +282,55 @@ def FeedForward(features: int, multiplier: int) -> nn.Module:
     )
 
 
-ConvT = T(Conv)
-ConvBlockT = T(ConvBlock)
-DownsampleT = T(Downsample)
-UpsampleT = T(Upsample)
-ResnetBlockT = T(ResnetBlock)
-AttentionT = T(Attention)
-FeedForwardT = T(FeedForward)
+def rand_bool(shape: Any, proba: float, device: Any = None) -> Tensor:
+    if proba == 1:
+        return torch.ones(shape, device=device, dtype=torch.bool)
+    elif proba == 0:
+        return torch.zeros(shape, device=device, dtype=torch.bool)
+    else:
+        return torch.bernoulli(torch.full(shape, proba, device=device)).to(torch.bool)
+
+
+def CFG(
+    net_t: Type[nn.Module],
+    embedding_max_length: int,
+) -> Callable[..., nn.Module]:
+    """Classifier-Free Guidance -> CFG(UNet, embedding_max_length=512)(...)"""
+
+    def CFGNet(embedding_features: int, **kwargs) -> nn.Module:
+        fixed_embedding = FixedEmbedding(
+            max_length=embedding_max_length,
+            features=embedding_features,
+        )
+        net = net_t(embedding_features=embedding_features, **kwargs)  # type: ignore
+
+        def forward(
+            x: Tensor,
+            embedding: Optional[Tensor] = None,
+            embedding_scale: float = 1.0,
+            embedding_mask_proba: float = 0.0,
+            **kwargs,
+        ):
+            assert exists(embedding), "embedding required when using CFG"
+            b, device = embedding.shape[0], embedding.device
+            embedding_mask = fixed_embedding(embedding)
+
+            if embedding_mask_proba > 0.0:
+                # Randomly mask embedding
+                batch_mask = rand_bool(
+                    shape=(b, 1, 1), proba=embedding_mask_proba, device=device
+                )
+                embedding = torch.where(batch_mask, embedding_mask, embedding)
+
+            if embedding_scale != 1.0:
+                # Compute both normal and fixed embedding outputs
+                out = net(x, embedding=embedding, **kwargs)
+                out_masked = net(x, embedding=embedding_mask, **kwargs)
+                # Scale conditional output using classifier-free guidance
+                return out_masked + (out - out_masked) * embedding_scale
+            else:
+                return net(x, embedding=embedding, **kwargs)
+
+        return Module([fixed_embedding, net], forward)
+
+    return CFGNet
